# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/ipfssearchw.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_IPFSSearchMain(object):
    def setupUi(self, IPFSSearchMain):
        IPFSSearchMain.setObjectName("IPFSSearchMain")
        IPFSSearchMain.resize(779, 463)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(IPFSSearchMain)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(20)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.comboPages = QtWidgets.QComboBox(IPFSSearchMain)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboPages.sizePolicy().hasHeightForWidth())
        self.comboPages.setSizePolicy(sizePolicy)
        self.comboPages.setMinimumSize(QtCore.QSize(140, 0))
        self.comboPages.setObjectName("comboPages")
        self.horizontalLayout.addWidget(self.comboPages)
        self.prevPageButton = QtWidgets.QPushButton(IPFSSearchMain)
        self.prevPageButton.setMinimumSize(QtCore.QSize(40, 0))
        self.prevPageButton.setText("")
        self.prevPageButton.setObjectName("prevPageButton")
        self.horizontalLayout.addWidget(self.prevPageButton)
        self.nextPageButton = QtWidgets.QPushButton(IPFSSearchMain)
        self.nextPageButton.setMinimumSize(QtCore.QSize(40, 0))
        self.nextPageButton.setText("")
        self.nextPageButton.setObjectName("nextPageButton")
        self.horizontalLayout.addWidget(self.nextPageButton)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.labelInfo = QtWidgets.QLabel(IPFSSearchMain)
        self.labelInfo.setText("")
        self.labelInfo.setObjectName("labelInfo")
        self.horizontalLayout.addWidget(self.labelInfo)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setHorizontalSpacing(12)
        self.gridLayout.setObjectName("gridLayout")
        self.itemTypeFilter = QtWidgets.QComboBox(IPFSSearchMain)
        self.itemTypeFilter.setObjectName("itemTypeFilter")
        self.gridLayout.addWidget(self.itemTypeFilter, 1, 1, 1, 1)
        self.contentTypeFilter = QtWidgets.QComboBox(IPFSSearchMain)
        self.contentTypeFilter.setObjectName("contentTypeFilter")
        self.gridLayout.addWidget(self.contentTypeFilter, 1, 3, 1, 1)
        self.lastSeenLabel = QtWidgets.QLabel(IPFSSearchMain)
        self.lastSeenLabel.setText("")
        self.lastSeenLabel.setObjectName("lastSeenLabel")
        self.gridLayout.addWidget(self.lastSeenLabel, 1, 6, 1, 1)
        self.label_4 = QtWidgets.QLabel(IPFSSearchMain)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 1, 0, 1, 1)
        self.label_5 = QtWidgets.QLabel(IPFSSearchMain)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 1, 4, 1, 1)
        self.lastSeenSlider = QtWidgets.QSlider(IPFSSearchMain)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lastSeenSlider.sizePolicy().hasHeightForWidth())
        self.lastSeenSlider.setSizePolicy(sizePolicy)
        self.lastSeenSlider.setMaximumSize(QtCore.QSize(300, 16777215))
        self.lastSeenSlider.setOrientation(QtCore.Qt.Horizontal)
        self.lastSeenSlider.setObjectName("lastSeenSlider")
        self.gridLayout.addWidget(self.lastSeenSlider, 1, 5, 1, 1)
        self.label_3 = QtWidgets.QLabel(IPFSSearchMain)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 0, 2, 1, 1)
        self.searchQuery = QtWidgets.QLineEdit(IPFSSearchMain)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.searchQuery.sizePolicy().hasHeightForWidth())
        self.searchQuery.setSizePolicy(sizePolicy)
        self.searchQuery.setObjectName("searchQuery")
        self.gridLayout.addWidget(self.searchQuery, 0, 1, 1, 1)
        self.label = QtWidgets.QLabel(IPFSSearchMain)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 2, 1, 1)
        self.filenameFilter = QtWidgets.QLineEdit(IPFSSearchMain)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filenameFilter.sizePolicy().hasHeightForWidth())
        self.filenameFilter.setSizePolicy(sizePolicy)
        self.filenameFilter.setObjectName("filenameFilter")
        self.gridLayout.addWidget(self.filenameFilter, 0, 3, 1, 1)
        self.label_2 = QtWidgets.QLabel(IPFSSearchMain)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 0, 0, 1, 1)
        self.label_6 = QtWidgets.QLabel(IPFSSearchMain)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 4, 1, 1)
        self.titleFilter = QtWidgets.QLineEdit(IPFSSearchMain)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.titleFilter.sizePolicy().hasHeightForWidth())
        self.titleFilter.setSizePolicy(sizePolicy)
        self.titleFilter.setMaximumSize(QtCore.QSize(200, 16777215))
        self.titleFilter.setObjectName("titleFilter")
        self.gridLayout.addWidget(self.titleFilter, 0, 5, 1, 1)
        self.label_7 = QtWidgets.QLabel(IPFSSearchMain)
        self.label_7.setEnabled(False)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 0, 6, 1, 1)
        self.resultsPerPage = QtWidgets.QComboBox(IPFSSearchMain)
        self.resultsPerPage.setEnabled(False)
        self.resultsPerPage.setObjectName("resultsPerPage")
        self.gridLayout.addWidget(self.resultsPerPage, 0, 7, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        self.browser = QtWebEngineWidgets.QWebEngineView(IPFSSearchMain)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.browser.sizePolicy().hasHeightForWidth())
        self.browser.setSizePolicy(sizePolicy)
        self.browser.setAutoFillBackground(False)
        self.browser.setStyleSheet("border-color: qlineargradient(spread:pad, x1:0, y1:0, x2:1, y2:0, stop:0 rgba(255, 178, 102, 255), stop:0.55 rgba(235, 148, 61, 255), stop:0.98 rgba(0, 0, 0, 255), stop:1 rgba(0, 0, 0, 0));\n"
"background-color: qradialgradient(spread:pad, cx:0.5, cy:0.5, radius:0.5, fx:0.5, fy:0.5, stop:0 rgba(255, 0, 0, 255), stop:0.479904 rgba(255, 0, 0, 255), stop:0.522685 rgba(255, 255, 255, 255), stop:1 rgba(255, 255, 255, 255));")
        self.browser.setUrl(QtCore.QUrl("about:blank"))
        self.browser.setObjectName("browser")
        self.verticalLayout_2.addWidget(self.browser)

        self.retranslateUi(IPFSSearchMain)
        QtCore.QMetaObject.connectSlotsByName(IPFSSearchMain)

    def retranslateUi(self, IPFSSearchMain):
        _translate = QtCore.QCoreApplication.translate
        IPFSSearchMain.setWindowTitle(_translate("IPFSSearchMain", "Form"))
        self.label_4.setText(_translate("IPFSSearchMain", "File type"))
        self.label_5.setText(_translate("IPFSSearchMain", "Last seen"))
        self.label_3.setText(_translate("IPFSSearchMain", "Filename search"))
        self.label.setText(_translate("IPFSSearchMain", "Content type"))
        self.label_2.setText(_translate("IPFSSearchMain", "Search"))
        self.label_6.setText(_translate("IPFSSearchMain", "Title"))
        self.label_7.setText(_translate("IPFSSearchMain", "Results per page"))

from PyQt5 import QtWebEngineWidgets
